# -*- coding: utf-8 -*-
"""Constant quantities of widely-accepted physical properties."""

from aguaclara.core.units import unit_registry as u

GRAVITY = 9.80665 * u.m / u.s ** 2

WATER_DENSITY = 1000 * u.kg / u.m ** 3
WATER_NU = 1 * 10 ** -6 * u.m ** 2 / u.s

ATM_P = 1 * u.atm
AIR_NU = 12 * u.mm ** 2 / u.s
AIR_DENSITY = 1.204 * u.kg / u.mm ** 3

# The influence of viscosity on mixing in jet reactors
JET_ROUND_RATIO = 0.08
# Estimate for plane jets in the flocculator and sed tank jet reverser.
JET_PLANE_RATIO = 0.0124

LFP_FLOW_MAX = 16.1 * u.L / u.s

FITTING_S_MIN = 5 * u.cm  # Between fittings and tank wall in a tank.
CHANNEL_W_MIN = 15 * u.cm

VC_ORIFICE_RATIO = 0.63

K_KOZENY = 5



from aguaclara.core.units import unit_registry as u

L_MAX = 2.2 * u.m

# Angle of the sloped walls of the entrance tank hoppers
ENT_TANK_SLOPE_ANGLE = 45 * u.deg

# Extra space around the float (increase in effective diameter) to
# ensure that float has free travel
FLOAT_S = 5 * u.cm
HOPPER_PEAK_W = 3 * u.cm
PLATE_S = 2.5 * u.cm
PLATE_THICKNESS = 2 * u.mm
CENTER_PLATE_DIST = PLATE_S + PLATE_THICKNESS
MOD_ND = 0.5 * u.inch

# Distance from the front wall to the pipe stubs in the hopper drains so
# that an operator can easily reach them.
WALL_DRAIN_DIST_MAX = 40 * u.cm

# Entrance tank capture velocity
CAPTURE_BOD_VEL = 8 * u.mm / u.s
PLATE_ANGLE = 50 * u.deg
MOD_SPACER_ND = 0.75 * u.inch

# Thickness of the PVC disk used as the float for the chemical dose
# controller lever arm.
FLOAT_THICKNESS = 5 * u.cm
LAMINA_PIPE_EDGE_S = 5 * u.cm

# Nom diam of the pipes that are embedded in the entrance tank slope
# to support the plate settler module
PLATE_SUPPORT_ND = 3 * u.inch


# Increased to get better mixing (10/10/2015 by Monroe)
RAPID_MIX_EDR = 3 * u.W / u.kg

RAPID_MIX_PLATE_RESTRAINER_ND = 0.5 * u.inch

FLOAT_ND = 8*u.inch

# Minimum pipe size to handle grit and to ensure that the pipe can be
# easily unclogged
DRAIN_MIN_ND = 3*u.inch

DRAIN_ND = 3*u.inch  # This is constant for now

REMOVABLE_WALL_THICKNESS = 5*u.cm

# Parameters are arbitrary - need to be calculated
REMOVABLE_WALL_SUPPORT_H = 4 * u.cm

REMOVABLE_WALL_SUPPORT_THICKNESS = 5*u.cm

HOPPER_LEDGE_THICKNESS = 15*u.cm
WALKWAY_W = 1 * u.m
RAPID_MIX_ORIFICE_PLATE_THICKNESS = 2*u.cm
RAPID_MIX_AIR_RELEASE_ND = 1*u.inch



from core.units import unit_registry as u

WALL_THICKNESS = 5*u.mm

# each element in the following array is tank volume

VOL_SUPPLIER = [208.198, 450, 600, 750, 1100, 2500]*u.L

# the following array is a 2D array in which
# in each element, the first element is tank diameter
# and the second element is tank height

DIMENSIONS_SUPPLIER = [[0.571, 0.851], [0.85, 0.99], [0.96, 1.10], [1.10, 1.02], [1.10, 1.39], [1.55, 1.65]]*u.m

FACTOR = [1.05, 1.05, 1.05, 1.05, 1.05, 1.05]


#-*- coding: utf-8 -*-
"""
Data for materials that are used in the construction of AguaClara water
treatment plants.
"""
from aguaclara.core.units import unit_registry as u

PVC_PIPE_ROUGH = 0.12 * u.mm

CONCRETE_PIPE_ROUGH = 2 * u.mm
CONCRETE_DENSITY = 2400 * (u.kg / u.m ** 3)
CONCRETE_THICKNESS_MIN = 5 * u.cm

REBAR_D = 0.5 * u.inch
ACRYLIC_T = 1 * u.cm



"""This module is intended to provide easy set up for a playground development
environment using the aguaclara package.

It imports all commonly used packages with one line, ensures Python is
run in the correct virtual environment, sets sig figs correctly and provides
any additional environment massaging to get to designing as quickly as
possible. This should NOT be used by other modules within aguaclara as it
results in unnecessary imports.

Usage:

    * Import all into your global namespace with: `from aguaclara.play *`
    * setup_aide() should run during import.

Now you should be able to execute:
    *`np.array([1,2,3,4])
And your numbers should be limited to four significant figures when printed.

"""

# Third-party imports
import math
import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt

# Design imports
# from aguaclara.design.lfom import LFOM    #NOTE: temporarily disabled for release 0.0.15 to prevent problems
                                            #with lacking Onshape set up (unnecessary for research teams)


# Core imports
import aguaclara
import aguaclara.core.pipes as pipe
from aguaclara.core.units import unit_registry as u
from aguaclara.core import physchem as pc
import aguaclara.core.constants as con
import aguaclara.core.materials as mat
import aguaclara.core.utility as ut
import aguaclara.core.head_loss as k
import aguaclara.core.pipeline as pipeline
# import aguaclara.core.optional_inputs as opt
import warnings


# deprecated imports
# import aguaclara.core.expert_inputs as exp


def setup_aide():
    """
    This is the public function that should be called to completely setup the aide environment
    in a jupyter notebook.
    :return:
    """
    matplotlib.style.use('ggplot')
    set_sig_fig()
    #ensure_in_a_virtual_environment()
        # Don't want to scare the kiddos with virtual environments yet! TODO.

def set_sig_fig(n: int = 4):
    """Set the default number of significant figures used to print pint, pandas and numpy values
    quantities. Defaults to 4.

    Args:
        n: number of significant figures to display

    Example:
        import aguaclara
        from aguaclara.units import unit_registry as u
        h=2.5532532522352543*u.m
        e = 25532532522352543*u.m
        print('h before sigfig adjustment: ',h)
        print('e before sigfig adjustment: ',e)
        aguaclara.units.set_sig_figs(10)
        print('h after sigfig adjustment: ',h)
        print('e after sigfig adjustment: ',e)

        h before sigfig adjustment:  2.553 meter
        e before sigfig adjustment:  2.553e+16 meter
        h after sigfig adjustment:  2.553253252 meter
        e after sigfig adjustment:  2.553253252e+16 meter
    """
    u.default_format = '.' + str(n) + 'g'
    pd.options.display.float_format = ('{:,.' + str(n) + '}').format


def ensure_in_a_virtual_environment():
    """
    Warn users if not in a virtual environment
    """
    import sys
    # way to test for virtual environment: https://stackoverflow.com/a/1883251/5136799
    if hasattr(sys, 'real_prefix'):
        raise UserWarning("aguaclara should always be run in a virtual environment to ensure"
                          "that all dependencies are correctly installed. Please refer to the"
                          "readme for virtual environment setup instructions.")

# Run when imported
setup_aide()



from aguaclara.core.units import unit_registry as u
import aguaclara.core.constants as con
import aguaclara.core.pipes as pipe

import numpy as np

PLANT_FLOOR_THICKNESS = 0.2 * u.m  # plant floor slab thickness
WALL_THICKNESS = 0.15 * u.m  # thickness of sed tank dividing wall
HL_OUTLET_MAN = 4 * u.cm  # head loss through the outlet manifold

BOD_UP_VEL = 1 * u.mm / u.s

##Plate settler
CONC_BOD_VEL = 0.12 * u.mm / u.s  # capture velocity

PLATE_ANGLE = 60 * u.deg

PLATE_S = 2.5 * u.cm

MODULE_PLATES_N_MIN = 8

# This is moved to template because SED_PLATE_THICKNESS is in materials.yaml
# CENTER_SED_PLATE_DIST = PLATE_S + SED_PLATE_THICKNESS

# Bottom of channel
SLOPE_ANGLE = 50 * u.deg

##This slope needs to be verified for functionality in the field.
# A steeper slope may be required in the floc hopper.
HOPPER_SLOPE_ANGLE = 45 * u.deg

WATER_H_EST = 2 * u.m

GATE_VALVE_URL = "https://confluence.cornell.edu/download/attachments/173604905/Sed-Scaled-Gate-Valve-Threaded.dwg"

SUPPORT_BOLT_URL = "https://confluence.cornell.edu/download/attachments/173604905/PlateSettlerSupportBolt.dwg"

##Inlet channel
WEIR_HL_MAX = 5 * u.cm

##Height of the inlet channel overflow weir above the normal water level
# in the inlet channel so that the far side of the overflow weir does not
# fill with water under normal operating conditions. This means the water
# level in the inlet channel will increase when the inlet overflow weir
# is in use.
INLET_WEIR_FREE_BOARD_H = 2 * u.cm

WEIR_THICKNESS = 5*u.cm

INLET_HL_MAX = 1 * u.cm

# ratio of the height to the width of the sedimentation tank inlet channel.
INLET_H_W_RATIO = 0.95

##Exit launder
##Target headloss through the launder orifices
LAUNDER_BOD_HL = 4 * u.cm

##Acceptable ratio of min to max flow through the launder orifices
FLOW_LAUNDER_ORIFICES_RATIO = 0.80

##Center to center spacing of orifices in the launder
CENTER_LAUNDER_EST_DIST = 10 * u.cm

##The additional length needed in the launder cap pipe that is to be
# inserted into the launder coupling
LAUNDER_CAP_EXCESS_L = 3 * u.cm

##Space between the top of the plate settlers and the bottom of the
# launder pipe
LAMELLA_TO_LAUNDER_H = 5 * u.cm

##The additional length needed in the launder cap pipe that is to be
# inserted into the launder coupling

##Diameter of the pipe used to hold the plate settlers together
MOD_ND = 0.5 * u.inch

##Diameter of the pipe used to create spacers. The spacers slide over the
# 1/2" pipe and are between the plates
MOD_SPACER_ND = 0.75 * u.inch

MOD_SPACER_SDR = 17

##This is the vertical thickness of the lip where the lamella support sits. mrf222
LAMELLA_LEDGE_THICKNESS = 8 * u.cm

LAMELLA_PIPE_EDGE_S = 5 * u.cm

##Approximate x-dimension spacing between cross pipes in the plate settler
# support frame.
CENTER_PLATE_FRAME_CROSS_DIST_EST = 0.8 * u.m

##Estimated plate length used to get an initial estimate of sedimentation
# tank active length.
PLATE_L_EST = 60 * u.cm

##Pipe size of the support frame that holds up the plate settler modules
PLATE_FRAME_ND = 1.5 * u.inch

##Floc weir

#Vertical distance from the top of the floc weir to the bottom of the pipe
# frame that holds up the plate settler modules
FLOC_WEIR_TO_PLATE_FRAME_H = 10 * u.cm

##Minimum length (X dimension) of the floc hopper
HOPPER_MIN_L = 50 * u.cm

##Inlet manifold
##Max energy dissipation rate in the sed diffuser outletS
ENERGY_DIS_INT_MAX = 150 * u.mW/u.kg

##Ratio of min to max flow through the inlet manifold diffusers
INLET_Q_RATIO = 0.8

MAN_ND_MAX = 8 * u.inch

MAN_SDR = 41  # SDR of pipe for sed tank inlet manifold

##This is the minimum distance between the inlet manifold and the slope
# of the sed tank.
INLET_MAN_SLOPE_S = 10 * u.cm

##Length of exposed manifold stub coming out of the floc weir to which the
# free portion of the inlet manifold is attached with a flexible coupling.
MAN_CONNECTION_STUB_L = 4 * u.cm

##Space between the end of the manifold pipe and the edge of the first
# diffuser's hole, or the first manifold orifice.

MAN_FIRST_DIFFUSER_GAP_L = 3 * u.cm

##Vertical distance from the edge of the jet reverser half-pipe to the tip
# of the inlet manifold diffusers
JET_REVERSER_TO_DIFFUSERS_H = 3 * u.cm

##Gap between the end of the inlet manifold pipe and the end wall of the
# tank to be able to install the pipe
MAN_PIPE_FROM_TANK_END_L = 2  *u.cm

WALL_TO_DIFFUSER_GAP_L_MIN = 3 * u.cm

# Diameter of the holes drilled in the manifold so that the molded 1"
# diffuser pipes can fit tightly in place (normal OD of a 1" pipe is
# close to 1-5/16")
MAN_PORT_D = 1.25 * u.inch

# nominal diameter of pipe used for jet reverser in bottom of set tank
JET_REVERSER_ND = 3 * u.inch

SDR_REVERSER = 26  # SDR of jet reverser pipe

# Diffuser geometry
SDR_DIFFUSER = 26  # SDR of diffuser pipe

DIFFUSER_PIPE_ND = 4 * u.cm  # nominal diameter of pipe used to make diffusers

AREA_PVC_DIFFUSER = (
    (np.pi/4) * ((pipe.OD(DIFFUSER_PIPE_ND)**2)
    - (pipe.ID_SDR(DIFFUSER_PIPE_ND, SDR_DIFFUSER))**2)
)
# stretch factor applied to the diffuser PVC pipes as they are heated
# and molded
PVC_STRETCH_RATIO = 1.2

T_DIFFUSER = ((pipe.OD(DIFFUSER_PIPE_ND) -
                        pipe.ID_SDR(DIFFUSER_PIPE_ND, SDR_DIFFUSER))
                              / (2 * PVC_STRETCH_RATIO))

W_DIFFUSER_INNER = 0.3175 * u.cm  # opening width of diffusers

# Calculating using a minor loss equation with K = 1
V_SED_DIFFUSER_MAX = np.sqrt(2 * con.GRAVITY * INLET_HL_MAX).to(u.mm / u.s)

DIFFUSER_L = 15 * u.cm  # vertical length of diffuser

B_DIFFUSER = 5 * u.cm  # center to center spacing beteen diffusers

# Headloss through the diffusers to ensure uniform flow between sed tanks
DIFFUSER_HL = 0.001 * u.m

# Outlet to filter
# If the plant has two trains, the current design shows the exit channel
# continuing from one set of sed tanks into the filter inlet channel.
# The execution of this extended channel involves a few calculations.
FILTER_OUTLET_HL_MAX = 10 * u.cm

# Maximum length of sed plate sticking out past module pipes without any
# additional support. The goal is to prevent floppy modules that don't maintain
# constant distances between the plates

PLATE_CANTILEVERED_L = 20 * u.cm

HOPPER_DRAIN_ND = 1*u.inch

HOPPER_VIEWER_ND = 2*u.inch

HOPPER_SKIMMER_ND = 2*u.inch

# Diffusers/Jet Reverser

DIFFUSER_ND = 1*u.inch

JET_REVERSER_ND = 3*u.inch

"""This file contains all the functions needed to design a sedimentation tank
for an AguaClara plant.
Attributes
----------
thickness_wall : float
    Thickness of walls in the sedimentation unit process
plate_settlers : dict
    A dictionary containing variables relating to the plate settlers
    Attributes
    ----------
    angle : int
        Angle of plate settlers (relative to being completely horizontal)
    S : float
        Edge to edge distance between plates
    thickness : float
        Thickness of PVC sheet used to make plate settlers
    L_cantilevered : float
        Maximum length of sed plate sticking out past module pipes without any
        additional support. The goal is to prevent floppy modules that don't
        maintain constant distances between the plates
tank : dict
    A dictionary containing variables relating to the concrete portion of the
    sedimentation tank
    Attributes
    ----------
    W : float
        Width of the sedimentation tank. Based off of the width of the PVC
        sheet used to make plate settlers
    L : float
        Length of the sedimentation tank. Based off of the length of a manifold
        pipe
    vel_up : float
        Upflow velocity through a sedimentation tank used as basis of design
manifold : dict
    A dictionary containg variables relating to the inlet manifold,
    exit manifold, and diffusers
    Attributes
    ----------
    ratio_Q_orifice : float
        Acceptable ratio of min to max flow through the manifold orifices
    diffuser : dict
        A dictionary containing variables relating to the diffuser
        Attributes
        ----------
        thickness_wall : float
            Wall thickness of a diffuser
        vel_max : float
            Maximum velocity through a diffuser
        A : float
            Area of a diffuser when viewed down the length of the manifold
    exit_man : dict
        A dictionary containing variables relating to the exit manifold
        Attributes
        ----------
        hl_orifice : float
            Headloss through an orifice in the exit manifold
        N_orifices : int
            Number of orifices in the exit manifold
"""
from aide_design.play import*

# again we will change this to an important statment from the URL of  aide_template repo
sed_dict = {
            'thickness_wall': 0.15*u.m,
            'plate_settlers': {
                'angle': 60*u.deg, 'S': 2.5*u.cm,
                'thickness': 2*u.mm, 'L_cantilevered': 20*u.cm,
                },
            'tank': {
                'W': 42*u.inch, 'L': 5.8*u.m, 'vel_up': 1*u.mm/u.s
            },
            'manifold': {
                'ratio_Q_man_orifice': 0.8,
                'diffuser': {
                    'thickness_wall': 1.17*u.inch, 'vel_max': 442.9*u.mm/u.s,
                    'A': 0.419*u.inch**2
                },
                'exit_man': {
                    'hl_orifice': 4*u.cm, 'N_orifices': 58
                }
            }
}

@u.wraps(None, [None], False)
def n_sed_plates_max(sed_inputs=sed_dict):
    """Return the maximum possible number of plate settlers in a module given
    plate spacing, thickness, angle, and unsupported length of plate settler.
    Parameters
    ----------
    S_plate : float
        Edge to edge distance between plate settlers
    thickness_plate : float
        Thickness of PVC sheet used to make plate settlers
    L_sed_plate_cantilevered : float
        Maximum length of sed plate sticking out past module pipes without any
        additional support. The goal is to prevent floppy modules that don't
        maintain constant distances between the plates
    angle_plate : float
        Angle of plate settlers
    Returns
    -------
    int
        Maximum number of plates
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    B_plate = sed_inputs['plate_settlers']['S'] + sed_inputs['plate_settlers']['thickness']
    return math.floor((sed_inputs['plate_settlers']['L_cantilevered'].magnitude / B_plate.magnitude
                      * np.tan(sed_inputs['plate_settlers']['angle'].to(u.rad).magnitude)) + 1)

@u.wraps(u.inch, [None], False)
def w_diffuser_inner_min(sed_inputs=sed_dict):
    """Return the minimum inner width of each diffuser in the sedimentation tank.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations. Can be found in sed.yaml
    Returns
    -------
    float
        Minimum inner width of each diffuser in the sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return ((sed_inputs['tank']['vel_up'].to(u.inch/u.s).magnitude /
             sed_inputs['manifold']['diffuser']['vel_max'].to(u.inch/u.s).magnitude)
             * sed_inputs['tank']['W'])

@u.wraps(u.m, [None], False)
def w_diffuser_inner(sed_inputs=sed_dict):
    """Return the inner width of each diffuser in the sedimentation tank.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Inner width of each diffuser in the sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return ut.ceil_nearest(w_diffuser_inner_min(sed_inputs).magnitude,
                           (np.arange(1/16,1/4,1/16)*u.inch).magnitude)

@u.wraps(u.m, [None], False)
def w_diffuser_outer(sed_inputs=sed_dict):
    """Return the outer width of each diffuser in the sedimentation tank.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Outer width of each diffuser in the sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return (w_diffuser_inner_min(sed_inputs['tank']['W']) +
            (2 * sed_inputs['manifold']['diffuser']['thickness_wall'])).to(u.m).magnitude

@u.wraps(u.m, [None], False)
def L_diffuser_outer(sed_inputs=sed_dict):
    """Return the outer length of each diffuser in the sedimentation tank.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Outer length of each diffuser in the sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return ((sed_inputs['manifold']['diffuser']['A'] /
           (2 * sed_inputs['manifold']['diffuser']['thickness_wall']))
           - w_diffuser_inner(sed_inputs).to(u.inch)).to(u.m).magnitude

@u.wraps(u.m, [None], False)
def L_diffuser_inner(sed_inputs=sed_dict):
    """Return the inner length of each diffuser in the sedimentation tank.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Inner length of each diffuser in the sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return L_diffuser_outer(sed_inputs['tank']['W']) -
            (2 * (sed_inputs['manifold']['diffuser']['thickness_wall']).to(u.m)).magnitude)

@u.wraps(u.m**3/u.s, [None], False)
def q_diffuser(sed_inputs=sed_dict):
    """Return the flow through each diffuser.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Flow through each diffuser in the sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return (sed_inputs['tank']['vel_up'].to(u.m/u.s) *
             sed_inputs['tank']['W'].to(u.m) *
             L_diffuser_outer(sed_inputs)).magnitude

@u.wraps(u.m/u.s, [None], False)
def vel_sed_diffuser(sed_inputs=sed_dict):
    """Return the velocity through each diffuser.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Flow through each diffuser in the sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return (q_diffuser(sed_inputs).magnitude
            / (w_diffuser_inner(w_tank) * L_diffuser_inner(w_tank)).magnitude)

@u.wraps(u.m**3/u.s, [None], False)
def q_tank(sed_inputs=sed_dict):
    """Return the maximum flow through one sedimentation tank.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Maximum flow through one sedimentation tank
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    return (sed_inputs['tank']['L'] * sed_inputs['tank']['vel_up'].to(u.m/u.s) *
            sed_inputs['tank']['W'].to(u.m)).magnitude

@u.wraps(u.m/u.s, [None], False)
def vel_inlet_man_max(sed_inputs=sed_dict):
    """Return the maximum velocity through the manifold.
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Maximum velocity through the manifold.
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    vel_manifold_max = (sed_inputs['diffuser']['vel_max'].to(u.m/u.s).magnitude *
        sqrt(2*((1-(sed_inputs['manifold']['ratio_Q_man_orifice'])**2)) /
        (((sed_inputs['manifold']['ratio_Q_man_orifice'])**2)+1)))
    return vel_manifold_max

@u.wraps(None, [u.m**3/u.s, None], False)
def n_tanks(Q_plant, sed_inputs=sed_dict):
    """Return the number of sedimentation tanks required for a given flow rate.
    Parameters
    ----------
    Q_plant : float
        Total plant flow rate
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    int
        Number of sedimentation tanks required for a given flow rate.
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    q = q_tank(sed_inputs).magnitude
    return (int(np.ceil(Q_plant / q)))

@u.wraps(u.m, [u.m**3/u.s, None], False)
def L_channel(Q_plant, sed_inputs=sed_dict):
    """Return the length of the inlet and exit channels for the sedimentation tank.
    Parameters
    ----------
    Q_plant : float
        Total plant flow rate
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Length of the inlet and exit channels for the sedimentation tank.
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    n_tanks = n_tanks(Q_plant, sed_inputs)
    return ((n_tanks * sed_inputs['tank']['W']) + sed_inputs['thickness_wall'] +
            ((n_tanks-1) * sed_inputs['thickness_wall']))

@u.wraps(u.m, [u.m**3/u.s, u.degK, None], False)
@ut.list_handler
def ID_exit_man(Q_plant, temp, sed_inputs=sed_dict):
    """Return the inner diameter of the exit manifold by guessing an initial
    diameter then iterating through pipe flow calculations until the answer
    converges within 1%% error
    Parameters
    ----------
    Q_plant : float
        Total plant flow rate
    temp : float
        Design temperature
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Inner diameter of the exit manifold
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    #Inputs do not need to be checked here because they are checked by
    #functions this function calls.
    nu = pc.viscosity_dynamic(temp)
    hl = sed_input['manifold']['exit_man']['hl_orifice'].to(u.m)
    L = sed_ipnut['manifold']['tank']['L']
    N_orifices = sed_inputs['manifold']['exit_man']['N_orifices']
    K_minor = con.K_MINOR_PIPE_EXIT
    pipe_rough = mat.PIPE_ROUGH_PVC.to(u.m)

    D = max(diam_pipemajor(Q_plant, hl, L, nu, pipe_rough).magnitude,
                   diam_pipeminor(Q_plant, hl, K_minor).magnitude)
    err = 1.00
    while err > 0.01:
            D_prev = D
            f = pc.fric(Q_plant, D_prev, nu, pipe_rough)
            D = ((8*Q_plant**2 / pc.GRAVITY.magnitude * np.pi**2 * hl) *
                    (((f*L/D_prev + K_minor) * (1/3 * 1/) *
                    (1/3 + 1/(2 * N_orifices) + 1/(6 * N_orifices**2)))
                    / (1 - sed_inputs['manifold']['ratio_Q_orifice']**2)))**0.25
            err = abs(D_prev - D) / ((D + D_prev) / 2)
    return D

@u.wraps(u.m, [u.m**3/u.s, u.inch, None], False)
def D_exit_man_orifice(Q_plant, drill_bits, sed_inputs=sed_dict):
    """Return the diameter of the orifices in the exit manifold for the sedimentation tank.
    Parameters
    ----------
    Q_plant : float
        Total plant flow rate
    drill_bits : list
        List of possible drill bit sizes
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Diameter of the orifices in the exit manifold for the sedimentation tank.
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    Q_orifice = Q_plant/sed_input['exit_man']['N_orifices']
    D_orifice = np.sqrt(Q_orifice**4)/(np.pi * con.RATIO_VC_ORIFICE * np.sqrt(2 * pc.GRAVITY.magnitude * sed_input['exit_man']['hl_orifice'].magnitude))
    return ut.ceil_nearest(D_orifice, drill_bits)


@u.wraps(u.m, [u.m**3/u.s, u.inch, None], False)
def L_sed_plate(sed_inputs=sed_dict):
    """Return the length of a single plate in the plate settler module based on
    achieving the desired capture velocity
    Parameters
    ----------
    sed_inputs : dict
        A dictionary of all of the constant inputs needed for sedimentation tank
        calculations can be found in sed.yaml
    Returns
    -------
    float
        Length of a single plate
    Examples
    --------
    >>> from aide_design.play import*
    >>>
    """
    L_sed_plate = ((sed_input['plate_settlers']['S'] * ((sed_input['tank']['vel_up']/sed_input['plate_settlers']['vel_capture'])-1)
                  + sed_input['plate_settlers']['thickness'] * (sed_input['tank']['vel_up']/sed_input['plate_settlers']['vel_capture']))
                 / (np.sin(sed_input['plate_settlers']['angle']) * np.cos(sed_input['plate_settlers']['angle']))
                 ).to(u.m)
    return L_sed_plate


import aguaclara.core.physchem as pc
import aguaclara.design.human_access as ha
import aguaclara.core.constants as con
from aguaclara.core.units import unit_registry as u
from onshapepy import Part

import math
import numpy as np

# Unused constants - START \/

FREEBOARD = 10 * u.cm

# From slope peak to weir
FLOC_BLANKET_H = 0.25 * u.m

# Distance that the rapid mix coupling extends into the first floc channel
# so that the RM orifice place can be fixed in place.
COUPLING_EXT_L = 5 * u.cm

# The minor loss coefficient is 2. According to measurements at Agalteca
# and according to
# https://confluence.cornell.edu/display/AGUACLARA/PAHO+Water+Treatment+Publications
# (page 100 in chapter on flocculation)
OPTION_H = 0

# Increased both to provide a safety margin on flocculator head loss and
# to simultaneously scale back on the actual collision potential we are
# trying to achieve.
BAFFLE_MINOR_LOSS = 2.5

BAFFLE_SET_BACK_PLASTIC_S = 2 * u.cm

# Target flocculator collision potential basis of design

BOD_GT = 75 * u.m ** (2 / 3)

# Ratio of the width of the gap between the baffle and the wall and the
# spacing between the baffles.
BAFFLE_RATIO = 1

# Max energy dissipation rate in the flocculator, basis of design.
BOD_ENERGY_DISSIPATION_MAX = 10 * u.mW / u.kg

DRAIN_TIME = 15 * u.min

MOD_ND = 0.5 * u.inch

SPACER_ND = 0.75 * u.inch

MOD_EDGE_LAST_PIPE_S = 10 * u.cm

RM_RESTRAINER_ND = 0.5 * u.inch

# Height that the drain stub extends above the top of the flocculator wall
DRAIN_STUB_EXT_H = 20 * u.cm

MOD_PIPE_EDGE_S = 10 * u.cm

BAFFLE_THICKNESS = 2 * u.mm

BAFFLE_RIGID_H_THICKNESS = 15*u.cm  # Is this a height or a thickness?

# The piping size for the main part of the floc modules
MODULES_MAIN_ND = (1/2)*u.inch

# The diameter of the oversized cap used to assemble the floc modules
MODULES_LARGE_ND = 1.5*u.inch

# Unused constants - END /\


class Flocculator:
    """Calculates physical dimensions of an AguaClara flocculator.

    Constant instance attributes
    ----------------------------
    - BAFFLE_K (K or K_baffle): float
        - The minor loss coefficient of the flocculator baffles.
    - HL (h_L_floc): float * u.cm
        - The target head loss in the flocculator.
    - GT (G-bar-Theta): float
        - The target collision potential of particles in the flocculator.
    - END_WATER_H (H): float * u.m
        - The height of water at the end of the flocculator.
    - CHANNEL_N_MIN (n_{Min, channel}): int
        - The minimum number of flocculator channels.
    - HS_RATIO_MIN (Pi_{HS}): float
        - The minimum ratio between expansion height and baffle spacing
    - HS_RATIO_MAX (Pi_{HS}): float
        - The maximum ratio between expansion height and baffle spacing
    - CAD: Part
        - URL to the flocculator 3D model in Onshape
    """

    # Increased both to provide a safety margin on flocculator head loss and
    # to simultaneously scale back on the actual collision potential we are
    # trying to achieve.
    # Originally calculated to be 2.3 from the equations:
    # VC_BAFFLE_RATIO = con.VC_ORIFICE_RATIO**2
    # K_MINOR_FLOC_BAFFLE = (1/VC_BAFFLE_RATIO - 1)**2
    BAFFLE_K = 2.5
    HL = 40 * u.cm
    GT = 37000
    END_WATER_H = 2 * u.m
    CHANNEL_N_MIN = 2
    HS_RATIO_MIN = 3
    HS_RATIO_MAX = 6

    CAD = Part(
        'https://cad.onshape.com/documents/b4cfd328713460beeb3125ac/w/3928b5c91bb0a0be7858d99e/e/6f2eeada21e494cebb49515f'
    )

    def __init__(self, q=20 * u.L/u.s, temp=25 * u.degC,
                 sed_tank_l_max=6 * u.m):
        """Instantiate a Flocculator object, representing a real flocculator
        component.

        :param q: Flow rate of water through the flocculator.
        :type q: float * u.L/u.s
        :param temp: Water temperature of the flocculator.
        :type temp: float * u.degC
        :param sed_tank_l_max: Maximum length of the sedimentation tank, used
        to calculate the length of the adjacent flocculator.
        :type sed_tank_l_max: float * u.m
        :returns: object
        :rtype: Flocculator
        """
        self.q = q
        self.temp = temp
        self.sed_tank_l_max = sed_tank_l_max

    @property
    def vel_grad_avg(self):
        """Calculate the average velocity gradient (G-bar) of water flowing
        through the flocculator.

        :returns: Average velocity gradient (G-bar)
        :rtype: float * 1 / second
        """
        return ((con.GRAVITY * self.HL) /
                (pc.viscosity_kinematic(self.temp) * self.GT)).to(u.s ** -1)

    @property
    def retention_time(self):
        """Calculate the retention time of flocs in a flocculator.

        :returns: Retention time of flocs (:math:`\theta`)
        :rtype: float * second
        """
        return (self.GT / self.vel_grad_avg).to(u.s)

    @property
    def vol(self):
        """Calculate the target volume of the flocculator.

        :returns: Target volume
        :rtype: float * meter ** 3
        """
        return (self.q * self.retention_time).to(u.m ** 3)

    @property
    def l_max_vol(self):
        """Calculate the maximum flocculator channel length that achieves the
        target volume, while still allowing human access.

        :returns: Maximum length based off of volume
        :rtype: float * meter
        """
        return (self.vol /
                (self.CHANNEL_N_MIN * ha.HUMAN_W_MIN * self.END_WATER_H)
                ).to(u.m)

    @property
    def channel_l(self):
        """Calculate the length of the flocculator channel that allows for the
        target volume, while at the same time, allowing for human access.

        :returns: Channel length
        :rtype: float * meter
        """
        return (min(self.sed_tank_l_max, self.l_max_vol)).to(u.m)

    @property
    def w_min_hs_ratio(self):
        """Calculate the minimum flocculator channel width, given the minimum
        ratio between expansion height (H) and baffle spacing (S).

        :returns: Minimum channel width given H_e/S
        :rtype: float * centimeter
        """
        return ((self.HS_RATIO_MIN * self.q.to(u.m ** 3 / u.s) / self.END_WATER_H) *
                (self.BAFFLE_K /
                 (2 * self.END_WATER_H * pc.viscosity_kinematic(self.temp) * self.vel_grad_avg ** 2)) ** (1/3)
                ).to(u.cm)

    @property
    def w_min(self):
        """Calculate the minimum channel width required to remain within the
        H_e/S ratio range and human access requirements.

        :returns: Minimum channel width
        :rtype: float * centimeter
        """
        return max(self.w_min_hs_ratio, ha.HUMAN_W_MIN).to(u.cm)

    @property
    def channel_n(self):
        """Return the number of channels in the entrance tank/flocculator (ETF).

        This takes the total width of the flocculator and divides it by the
        minimum channel width. A floor function is used to ensure that there
        are an even number of channels.

        :returns: Number of channels
        :rtype: int
        """
        num = self.w_total / self.w_min
        # floor function with step size 2
        num_floor = np.floor(num / 2) * 2
        return int(max(num_floor, 2))

    @property
    def w_total(self):
        """The total width of the flocculator.

        :returns: Total width
        :rtype: float * meter
        """
        return (self.vol / (self.channel_l * self.END_WATER_H)).to(u.m)

    @property
    def channel_w(self):
        """
        The channel width of the flocculator.  See section 'Flocculation
        Design' of textbook'

        :returns: Channel width
        :rtype: float * meter
        """
        return (self.w_total / self.channel_n).to(u.m)

    @property
    def expansion_h_max(self):
        """"Return the maximum distance between expansions for the largest
        allowable H/S ratio.

        :returns: Maximum expansion distance
        :rtype: float * meter

        Examples
        --------
        exp_dist_max(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)
        0.375 meter
        """
        return (((self.BAFFLE_K / (2 * pc.viscosity_kinematic(self.temp) * (self.vel_grad_avg ** 2))) *
                 (self.q * self.HS_RATIO_MAX / self.channel_w) ** 3) ** (1/4)).to(u.m)

    @property
    def expansion_n(self):
        """Return the minimum number of expansions per baffle space.

        :returns: Minimum number of expansions/baffle space
        :rtype: int
        """
        return math.ceil(self.END_WATER_H / self.expansion_h_max)

    @property
    def expansion_h(self):
        """Returns the height between flow expansions.

        :returns: Height between flow expansions
        :rtype: float * centimeter
        """
        return (self.END_WATER_H / self.expansion_n).to(u.cm)

    @property
    def baffle_s(self):
        """Return the spacing between baffles.

        :returns: Minimum number of expansions/baffle space
        :rtype: int
        """

        return ((self.BAFFLE_K /
                 (2 * self.expansion_h_max * (self.vel_grad_avg ** 2) *
                  pc.viscosity_kinematic(self.temp))) ** (1/3) *
                self.q / ha.HUMAN_W_MIN).to(u.cm)

    @property
    def obstacle_n(self):
        """Return the number of baffles a channel can contain.

        :returns: Number of baffles channel can contain
        :rtype: int
        """
        return (self.END_WATER_H / self.expansion_h) - 1

    def draw(self):
        """Draw the Onshape flocculator model based off of this object."""
        self.CAD.params = {
            'channel_l': self.channel_l,
            'channel_w': self.channel_w,
            'channel_h': self.END_WATER_H,
            'channel_pairs': self.channel_n,
            'baffle_s': self.baffle_s,
        }



import unittest
import os

import pandas as pd
import numpy as np

from aguaclara.core.units import unit_registry as u
from aguaclara.core import pipes

# TODO: Rewrite this testing class.
# class TestPipes(unittest.TestCase):
#
#     def test_pipes(self):
#         pipe = pipes.Pipe(nd=(7.0 * u.inch), sdr=35.0)
#         pipe_df = pd.read_csv(os.path.join(
#             os.path.dirname(__file__),
#             '../../aguaclara/core/data/pipe_database.csv'
#         ))
#         assert_list_almost_equal = np.vectorize(self.assertAlmostEqual)
#         assert_list_almost_equal(
#             [
#                 pipe.od,
#                 pipe.id_sdr,
#                 pipe.id_sch40,
#             ],
#             [
#                 7.625 * u.inch,
#                 7.1892857 * u.inch,
#                 7.023 * u.inch,
#             ]
#         )
#         assert_list_almost_equal(
#             [
#                 pipes.ND_all_available(),
#                 pipes.ID_SDR_all_available(35.0),
#                 pipes.ND_SDR_available(7.1892857 * u.inch, 35.0),
#                 pipes.ND_available(4.7 * u.inch)
#             ],
#             [
#                 (pipe_df[pipe_df['Used'] == 1]['NDinch'].values) * u.inch,
#                 (pipe_df['NDinch'].values * 0.1142856) * u.inch,
#                 7.0 * u.inch,
#                 6.0 * u.inch
#             ]
#         )



from aguaclara.core.units import unit_registry as u
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import os
from pathlib import Path


def get_data_by_time(path, columns, start_date, start_time="00:00", end_date=None, end_time="23:59"):
    """Extracts columns of data from a ProCoDA datalog based on starting and ending date(s) and times

    Note: currently only works for 1 or 2 days of data, i.e. end_date must be unspecified or one day after start_date

    Parameters
    ----------
    path : string
        The path to the folder containing your ProCoDA data files
    columns : int (list)
        A single index of a column or a list of indices of columns of data to extract
        Note: Column 0 is time. The first data column is column 1.
    start_date : string
        Starting date of data to extract, formatted 'M-D-YYYY'
    start_time: string, optional
        Starting time of data to extract, formatted 'HH:MM' (24-hour time)
    end_date : string, optional
        Ending date of data to extract, formatted 'M-D-YYYY'
    end_time: string, optional
        Ending time of data to extract, formatted 'HH:MM' (24-hour time)

    Return
    ------
    list (2D list)
        list :
            contains the single column of data to extract
        2D list:
            a list of lists containing the columns to extract, in order of the indices given in the columns variable

    Examples
    --------
    get_data_by_time(path='/Users/.../ProCoDA Data/', columns=4, start_date='6-14-2018', start_time='12:20',
        end_date='6-15-2018', end_time='10:50')
    get_data_by_time(path='/Users/.../ProCoDA Data/', columns=[0,4], start_date='6-14-2018', start_time='12:20',
         end_time='23:59')
    get_data_by_time(path='/Users/.../ProCoDA Data/', columns=[0,3,4], start_date='6-14-2018', end_date='6-18-2018')
    """

    # Locate and read data file(s)
    if path[-1] != '/':
        path += '/'
    paths = [path + "datalog " + start_date + '.xls']
    data = [remove_notes(pd.read_csv(paths[0], delimiter='\t'))]

    if end_date is not None:
        paths.append(path + "datalog " + end_date + ".xls")
        data.append(remove_notes(pd.read_csv(paths[1], delimiter='\t')))

    # Calculate start index
    time_column = pd.to_numeric(data[0].iloc[:, 0])
    interval = time_column[1]-time_column[0]
    start_idx = int(round((day_fraction(start_time) - time_column[0])/interval + .5)) #round up

    # Calculate end index
    time_column = pd.to_numeric(data[-1].iloc[:, 0])
    end_idx = int(round((day_fraction(end_time) - time_column[0])/interval + .5)) + 1 #round up

    # Get columns of interest
    if len(paths) == 1:
        if isinstance(columns, int):
            result = list(pd.to_numeric(data[0].iloc[start_idx:end_idx, columns]))
        else:
            result = []
            for c in columns:
                result.append(list(pd.to_numeric(data[0].iloc[start_idx:end_idx, c])))
    else:
        data[1].iloc[0, 0] = 0
        if isinstance(columns, int):
            result = list(pd.to_numeric(data[0].iloc[start_idx:, columns])) + \
                     list(pd.to_numeric(data[1].iloc[:end_idx, columns]) + (1 if columns == 0 else 0))
        else:
            result = []
            for c in columns:
                result.append(list(pd.to_numeric(data[0].iloc[start_idx:, c])) +
                              list(pd.to_numeric(data[1].iloc[:end_idx, c])+(1 if c == 0 else 0)))

    return result


def remove_notes(data):
    """Omits any rows containing text from a pandas.DataFrame object, except for headers

    Text is defined as characters of the alphabet. The resulting DataFrame should have only headers and numerical data.

    Parameters
    ----------
    data : pandas.DataFrame
        DataFrame object to remove text from

    Returns
    -------
    pandas.DataFrame
        DataFrame object with no text, except for headers
    """
    has_text = data.iloc[:, 0].astype(str).str.contains('(?!e-)[a-zA-Z]')
    text_rows = list(has_text.index[has_text])
    return data.drop(text_rows)


def day_fraction(time):
    """Converts a 24-hour time to a fraction of a day.

    For example, midnight corresponds to 0.0, and noon to 0.5.

    Parameters
    ----------
    time : string
        Time in the form of 'HH:MM' (24-hour time)

    Returns
    -------
    float
        A day fraction

    Examples
    --------
    >>> from aguaclara.research.procoda_parser import day_fraction
    >>> day_fraction("00:21")
    0.014583333333333334
    >>> day_fraction("18:30")
    0.7708333333333334
    """
    hour = int(time.split(":")[0])
    minute = int(time.split(":")[1])
    return hour/24 + minute/1440


def get_data_by_state(path, dates, state, column):
    """Reads a ProCoDA file and extracts the time and data column for each iteration of
    the given state.

    Parameters
    ----------
    dates : string (list)
        A list of dates or single date for which data was recorded, in
        the form "M-D-YYYY"
    state : int
        The state ID number for which data should be plotted
    column : int or string
        int:
            Index of the column that you want to extract. Column 0 is time.
            The first data column is column 1.
        string:
            Name of the column header that you want to extract
    path : string, optional
        Optional argument of the path to the folder containing your ProCoDA
        files. Defaults to the current directory if no argument is passed in
    extension : string, optional
        The file extension of the tab delimited file. Defaults to ".xls" if
        no argument is passed in

    Returns
    -------
    3-D list
        A list of lists of the time and data columns extracted for each iteration of the state.
         For example, if "data" is the output, data[i][:,0] gives the time column and data[i][:,1]
         gives the data column for the ith iteration of the given state and column. data[i][0]
         would give the first [time, data] pair.

    Examples
    --------
    get_data_by_state(["6-19-2013", "6-20-2013"], 1, 28)
    """
    data_agg = []
    day = 0
    first_day = True
    overnight = False
    extension = ".xls"
    if path[-1] != '/':
        path += '/'

    if not isinstance(dates, list):
        dates = [dates]

    for d in dates:
        state_file = path + "statelog " + d + extension
        data_file = path + "datalog " + d + extension

        states = pd.read_csv(state_file, delimiter='\t')
        data = pd.read_csv(data_file, delimiter='\t')

        states = np.array(states)
        data = np.array(data)

        # get the start and end times for the state
        state_start_idx = states[:, 1] == state
        state_start = states[state_start_idx, 0]
        state_end_idx = np.append([False], state_start_idx[0:-1])
        state_end = states[state_end_idx, 0]

        if overnight:
            state_start = np.insert(state_start, 0, 0)
            state_end = np.insert(state_end, 0, states[0, 0])

        if state_start_idx[-1]:
            np.append(state_end, data[0, -1])

        # get the corresponding indices in the data array
        data_start = []
        data_end = []
        for i in range(np.size(state_start)):
            add_start = True
            for j in range(np.size(data[:, 0])):
                if (data[j, 0] > state_start[i]) and add_start:
                    data_start.append(j)
                    add_start = False
                if data[j, 0] > state_end[i]:
                    data_end.append(j-1)
                    break

        if first_day:
            start_time = data[0, 0]

        # extract data at those times
        for i in range(np.size(data_start)):
            t = data[data_start[i]:data_end[i], 0] + day - start_time
            if isinstance(column, int):
                c = data[data_start[i]:data_end[i], column]
            else:
                c = data[column][data_start[i]:data_end[i]]
            if overnight and i == 0:
                data_agg = np.insert(data_agg[-1], np.size(data_agg[-1][:, 0]),
                                     np.vstack((t, c)).T)
            else:
                data_agg.append(np.vstack((t, c)).T)

        day += 1
        if first_day:
            first_day = False
        if state_start_idx[-1]:
            overnight = True

    return data_agg


def column_of_time(data_file_path, start, end=-1):
    """This function extracts the column of times from a ProCoDA data file.

    Parameters
    ----------
    data_file_path : string
        File path. If the file is in the working directory, then the file name
        is sufficient.

    start : int or float
        Index of first row of data to extract from the data file

    end : int or float, optional
        Index of last row of data to extract from the data
        Defaults to -1, which extracts all the data in the file

    Returns
    -------
    numpy array
        Experimental times starting at 0 day with units of days.

    Examples
    --------
    ftime(Reactor_data.txt, 0)

    """
    if not isinstance(start, int):
        start = int(start)
    if not isinstance(end, int):
        end = int(end)

    df = pd.read_csv(data_file_path, delimiter='\t')
    start_time = pd.to_numeric(df.iloc[start, 0])*u.day
    day_times = pd.to_numeric(df.iloc[start:end, 0])
    time_data = np.subtract((np.array(day_times)*u.day), start_time)
    return time_data


def column_of_data(data_file_path, start, column, end="-1", units=""):
    """This function extracts a column of data from a ProCoDA data file.

    Parameters
    ----------
    data_file_path : string
        File path. If the file is in the working directory, then the file name
        is sufficient.

    start : int
        Index of first row of data to extract from the data file

    end : int, optional
        Index of last row of data to extract from the data
        Defaults to -1, which extracts all the data in the file

    column : int or string
        int:
            Index of the column that you want to extract. Column 0 is time.
            The first data column is column 1.
        string:
            Name of the column header that you want to extract

    units : string, optional
        The units you want to apply to the data, e.g. 'mg/L'.
        Defaults to "" which indicates no units

    Returns
    -------
    numpy array
        Experimental data with the units applied.

    Examples
    --------
    column_of_data(Reactor_data.txt, 0, 1, -1, "mg/L")

    """
    if not isinstance(start, int):
        start = int(start)
    if not isinstance(end, int):
        end = int(end)

    df = pd.read_csv(data_file_path, delimiter='\t')
    if units == "":
        if isinstance(column, int):
            data = np.array(pd.to_numeric(df.iloc[start:end, column]))
        else:
            df[column][0:len(df)]
    else:
        if isinstance(column, int):
            data = np.array(pd.to_numeric(df.iloc[start:end, column]))*u(units)
        else:
            df[column][0:len(df)]*u(units)
    return data


def notes(data_file_path):
    """This function extracts any experimental notes from a ProCoDA data file.

    Parameters
    ----------
    data_file_path : string
        File path. If the file is in the working directory, then the file name
        is sufficient.

    Returns
    -------
    dataframe
        The rows of the data file that contain text notes inserted during the
        experiment. Use this to identify the section of the data file that you
        want to extract.

    Examples
    --------

    """
    df = pd.read_csv(data_file_path, delimiter='\t')
    text_row = df.iloc[0:-1, 0].str.contains('[a-z]', '[A-Z]')
    text_row_index = text_row.index[text_row].tolist()
    notes = df.loc[text_row_index]
    return notes


def read_state(dates, state, column, units="", path="", extension=".xls"):
    """Reads a ProCoDA file and outputs the data column and time vector for
    each iteration of the given state.

    Parameters
    ----------
    dates : string (list)
        A list of dates or single date for which data was recorded, in
        the form "M-D-Y"

    state : int
        The state ID number for which data should be extracted

    column : int or string
        int:
            Index of the column that you want to extract. Column 0 is time.
            The first data column is column 1.
        string:
            Name of the column header that you want to extract

    units : string, optional
        The units you want to apply to the data, e.g. 'mg/L'.
        Defaults to "" which indicates no units

    path : string, optional
        Optional argument of the path to the folder containing your ProCoDA
        files. Defaults to the current directory if no argument is passed in

    extension : string, optional
        The file extension of the tab delimited file. Defaults to ".xls" if
        no argument is passed in

    Returns
    -------
    time : numpy array
        Times corresponding to the data (with units)

    data : numpy array
        Data in the given column during the given state with units

    Examples
    --------
    time, data = read_state(["6-19-2013", "6-20-2013"], 1, 28, "mL/s")

    """
    data_agg = []
    day = 0
    first_day = True
    overnight = False

    if not isinstance(dates, list):
        dates = [dates]

    for d in dates:
        state_file = path + "statelog " + d + extension
        data_file = path + "datalog " + d + extension

        states = pd.read_csv(state_file, delimiter='\t')
        data = pd.read_csv(data_file, delimiter='\t')

        states = np.array(states)
        data = np.array(data)

        # get the start and end times for the state
        state_start_idx = states[:, 1] == state
        state_start = states[state_start_idx, 0]
        state_end_idx = np.append([False], state_start_idx[0:(np.size(state_start_idx)-1)])
        state_end = states[state_end_idx, 0]

        if overnight:
            state_start = np.insert(state_start, 0, 0)
            state_end = np.insert(state_end, 0, states[0, 0])

        if state_start_idx[-1]:
            state_end.append(data[0, -1])

        # get the corresponding indices in the data array
        data_start = []
        data_end = []
        for i in range(np.size(state_start)):
            add_start = True
            for j in range(np.size(data[:, 0])):
                if (data[j, 0] > state_start[i]) and add_start:
                    data_start.append(j)
                    add_start = False
                if (data[j, 0] > state_end[i]):
                    data_end.append(j-1)
                    break

        if first_day:
            start_time = data[1, 0]

        # extract data at those times
        for i in range(np.size(data_start)):
            t = data[data_start[i]:data_end[i], 0] + day - start_time
            if isinstance(column, int):
                c = data[data_start[i]:data_end[i], column]
            else:
                c = data[column][data_start[i]:data_end[i]]
            if overnight and i == 0:
                data_agg = np.insert(data_agg[-1], np.size(data_agg[-1][:, 0]),
                                     np.vstack((t, c)).T)
            else:
                data_agg.append(np.vstack((t, c)).T)

        day += 1
        if first_day:
            first_day = False
        if state_start_idx[-1]:
            overnight = True

    data_agg = np.vstack(data_agg)
    if units != "":
        return data_agg[:, 0]*u.day, data_agg[:, 1]*u(units)
    else:
        return data_agg[:, 0]*u.day, data_agg[:, 1]


def average_state(dates, state, column, units="", path="", extension=".xls"):
    """Outputs the average value of the data for each instance of a state in
    the given ProCoDA files

    Parameters
    ----------
    dates : string (list)
        A list of dates or single date for which data was recorded, in
        the form "M-D-Y"

    state : int
        The state ID number for which data should be extracted

    column : int or string
        int:
            Index of the column that you want to extract. Column 0 is time.
            The first data column is column 1.
        string:
            Name of the column header that you want to extract

    units : string, optional
        The units you want to apply to the data, e.g. 'mg/L'.
        Defaults to "" which indicates no units

    path : string, optional
        Optional argument of the path to the folder containing your ProCoDA
        files. Defaults to the current directory if no argument is passed in

    extension : string, optional
        The file extension of the tab delimited file. Defaults to ".xls" if
        no argument is passed in

    Returns
    -------
    float list
        A list of averages for each instance of the given state

    Examples
    --------
    data_avgs = average_state(["6-19-2013", "6-20-2013"], 1, 28, "mL/s")

    """
    data_agg = []
    day = 0
    first_day = True
    overnight = False

    if not isinstance(dates, list):
        dates = [dates]

    for d in dates:
        state_file = path + "statelog " + d + extension
        data_file = path + "datalog " + d + extension

        states = pd.read_csv(state_file, delimiter='\t')
        data = pd.read_csv(data_file, delimiter='\t')

        states = np.array(states)
        data = np.array(data)

        # get the start and end times for the state
        state_start_idx = states[:, 1] == state
        state_start = states[state_start_idx, 0]
        state_end_idx = np.append([False], state_start_idx[0:(np.size(state_start_idx)-1)])
        state_end = states[state_end_idx, 0]

        if overnight:
            state_start = np.insert(state_start, 0, 0)
            state_end = np.insert(state_end, 0, states[0, 0])

        if state_start_idx[-1]:
            state_end.append(data[0, -1])

        # get the corresponding indices in the data array
        data_start = []
        data_end = []
        for i in range(np.size(state_start)):
            add_start = True
            for j in range(np.size(data[:, 0])):
                if (data[j, 0] > state_start[i]) and add_start:
                    data_start.append(j)
                    add_start = False
                if (data[j, 0] > state_end[i]):
                    data_end.append(j-1)
                    break

        if first_day:
            start_time = data[1, 0]

        # extract data at those times
        for i in range(np.size(data_start)):
            if isinstance(column, int):
                c = data[data_start[i]:data_end[i], column]
            else:
                c = data[column][data_start[i]:data_end[i]]
            if overnight and i == 0:
                data_agg = np.insert(data_agg[-1], np.size(data_agg[-1][:]), c)
            else:
                data_agg.append(c)

        day += 1
        if first_day:
            first_day = False
        if state_start_idx[-1]:
            overnight = True

    averages = np.zeros(np.size(data_agg))
    for i in range(np.size(data_agg)):
        averages[i] = np.average(data_agg[i])

    if units != "":
        return averages*u(units)
    else:
        return averages


def perform_function_on_state(func, dates, state, column, units="", path="", extension=".xls"):
    """Performs the function given on each state of the data for the given state
    in the given column and outputs the result for each instance of the state

    Parameters
    ----------
    func : function
        A function which will be applied to data from each instance of the state

    dates : string (list)
        A list of dates or single date for which data was recorded, in
        the form "M-D-Y"

    state : int
        The state ID number for which data should be extracted

    column : int or string
        int:
            Index of the column that you want to extract. Column 0 is time.
            The first data column is column 1.
        string:
            Name of the column header that you want to extract

    units : string, optional
        The units you want to apply to the data, e.g. 'mg/L'.
        Defaults to "" which indicates no units

    path : string, optional
        Optional argument of the path to the folder containing your ProCoDA
        files. Defaults to the current directory if no argument is passed in

    extension : string, optional
        The file extension of the tab delimited file. Defaults to ".xls" if
        no argument is passed in

    Returns
    -------
    list
        The outputs of the given function for each instance of the given state

    Requires
    --------
    func takes in a list of data with units and outputs the correct units

    Examples
    --------
    def avg_with_units(lst):
        num = np.size(lst)
        acc = 0
        for i in lst:
            acc = i + acc

        return acc / num

    data_avgs = perform_function_on_state(avg_with_units, ["6-19-2013", "6-20-2013"], 1, 28, "mL/s")

    """
    data_agg = []
    day = 0
    first_day = True
    overnight = False

    if not isinstance(dates, list):
        dates = [dates]

    for d in dates:
        state_file = path + "statelog " + d + extension
        data_file = path + "datalog " + d + extension

        states = pd.read_csv(state_file, delimiter='\t')
        data = pd.read_csv(data_file, delimiter='\t')

        states = np.array(states)
        data = np.array(data)

        # get the start and end times for the state
        state_start_idx = states[:, 1] == state
        state_start = states[state_start_idx, 0]
        state_end_idx = np.append([False], state_start_idx[0:(np.size(state_start_idx)-1)])
        state_end = states[state_end_idx, 0]

        if overnight:
            state_start = np.insert(state_start, 0, 0)
            state_end = np.insert(state_end, 0, states[0, 0])

        if state_start_idx[-1]:
            state_end.append(data[0, -1])

        # get the corresponding indices in the data array
        data_start = []
        data_end = []
        for i in range(np.size(state_start)):
            add_start = True
            for j in range(np.size(data[:, 0])):
                if (data[j, 0] > state_start[i]) and add_start:
                    data_start.append(j)
                    add_start = False
                if (data[j, 0] > state_end[i]):
                    data_end.append(j-1)
                    break

        if first_day:
            start_time = data[1, 0]

        # extract data at those times
        for i in range(np.size(data_start)):
            if isinstance(column, int):
                c = data[data_start[i]:data_end[i], column]
            else:
                c = data[column][data_start[i]:data_end[i]]
            if overnight and i == 0:
                data_agg = np.insert(data_agg[-1], np.size(data_agg[-1][:]), c)
            else:
                data_agg.append(c)

        day += 1
        if first_day:
            first_day = False
        if state_start_idx[-1]:
            overnight = True

    output = np.zeros(np.size(data_agg))
    for i in range(np.size(data_agg)):
        if units != "":
            output[i] = func(data_agg[i]*u(units)).magnitude
        else:
            output[i] = func(data_agg[i])

    if units != "":
        return output*func(data_agg[i]*u(units)).units
    else:
        return output


def read_state_with_metafile(func, state, column, path, metaids=[],
                             extension=".xls", units=""):
    """Takes in a ProCoDA meta file and performs a function for all data of a
    certain state in each of the experiments (denoted by file paths in then
    metafile)

    Parameters
    ----------
    func : function
        A function which will be applied to data from each instance of the state

    state : int
        The state ID number for which data should be extracted

    column : int or string
        int:
            Index of the column that you want to extract. Column 0 is time.
            The first data column is column 1.
        string:
            Name of the column header that you want to extract

    path : string
        Path to your ProCoDA metafile (must be tab-delimited)

    metaids : string list, optional
        a list of the experiment IDs you'd like to analyze from the metafile

    extension : string, optional
        The file extension of the tab delimited file. Defaults to ".xls" if
        no argument is passed in

    units : string, optional
        The units you want to apply to the data, e.g. 'mg/L'.
        Defaults to "" which indicates no units

    Returns
    -------
    ids : string list
        The list of experiment ids given in the metafile

    outputs : list
        The outputs of the given function for each experiment

    Examples
    --------
    def avg_with_units(lst):
        num = np.size(lst)
        acc = 0
        for i in lst:
            acc = i + acc

        return acc / num

    path = "../tests/data/Test Meta File.txt"
    ids, answer = read_state_with_metafile(avg_with_units, 1, 28, path, [], ".xls", "mg/L")

    """
    outputs = []

    metafile = pd.read_csv(path, delimiter='\t', header=None)
    metafile = np.array(metafile)

    ids = metafile[1:, 0]

    if not isinstance(ids[0], str):
        ids = list(map(str, ids))

    if metaids:
        paths = []
        for i in range(len(ids)):
            if ids[i] in metaids:
                paths.append(metafile[i, 4])
    else:
        paths = metafile[1:, 4]

    basepath = os.path.join(os.path.split(path)[0], metafile[0, 4])

    # use a loop to evaluate each experiment in the metafile
    for i in range(len(paths)):
        # get the range of dates for experiment i
        day1 = metafile[i+1, 1]

        # modify the metafile date so that it works with datetime format
        if not (day1[2] == "-" or day1[2] == "/"):
            day1 = "0" + day1
        if not (day1[5] == "-" or day1[5] == "/"):
            day1 = day1[:3] + "0" + day1[3:]

        if day1[2] == "-":
            dt = datetime.strptime(day1, "%m-%d-%Y")
        else:
            dt = datetime.strptime(day1, "%m/%d/%y")
        duration = metafile[i+1, 3]

        if not isinstance(duration, int):
            duration = int(duration)

        date_list = []
        for j in range(duration):
            curr_day = dt.strftime("%m-%d-%Y")
            if curr_day[3] == "0":
                curr_day = curr_day[:3] + curr_day[4:]
            if curr_day[0] == "0":
                curr_day = curr_day[1:]

            date_list.append(curr_day)

            dt = dt + timedelta(days=1)

        path = str(Path(os.path.join(basepath, paths[i]))) + os.sep
        _, data = read_state(date_list, state, column, units, path, extension)

        outputs.append(func(data))

    return ids, outputs


def write_calculations_to_csv(funcs, states, columns, path, headers, out_name,
                              metaids=[], extension=".xls"):
    """Writes each output of the given functions on the given states and data
    columns to a new column in a

    Parameters
    ----------
    funcs : function (list)
        A function or list of functions which will be applied in order to the
        data. If only one function is given it is applied to all the
        states/columns

    states : string (list)
        The state ID numbers for which data should be extracted. List should be
        in order of calculation or if only one state is given then it will be
        used for all the calculations

    columns : int or string (list)
        If only one column is given it is used for all the calculations
            int:
                Index of the column that you want to extract. Column 0 is time.
                The first data column is column 1.
            string:
                Name of the column header that you want to extract

    path : string
        Path to your ProCoDA metafile (must be tab-delimited)

    headers : string list
        List of the desired header for each calculation, in order

    out_name : string
        Desired name for the output file. Can include a relative path

    metaids : string list, optional
        a list of the experiment IDs you'd like to analyze from the metafile

    extension : string, optional
        The file extension of the tab delimited file. Defaults to ".xls" if
        no argument is passed in

    Returns
    -------
    out_name.csv
        A CSV file with the each column being a new calcuation and each row
        being a new experiment on which the calcuations were performed

    output : DataFrame
        Pandas dataframe which is the same data that was written to CSV

    Requires
    --------
    funcs, states, columns, and headers are all of the same length if they are
    lists. Some being lists and some single values are okay.

    Examples
    --------


    """
    if not isinstance(funcs, list):
        funcs = [funcs] * len(headers)

    if not isinstance(states, list):
        states = [states] * len(headers)

    if not isinstance(columns, list):
        columns = [columns] * len(headers)

    data_agg = []
    for i in range(len(headers)):
        ids, data = read_state_with_metafile(funcs[i], states[i], columns[i],
                                             path, metaids, extension)
        data_agg = np.append(data_agg, [data])

    output = pd.DataFrame(data=np.vstack((ids, data_agg)).T,
                          columns=["ID"]+headers)
    output.to_csv(out_name, sep='\t')

    return output



"""Build an Linear Flow Orifice Meter"""
import aguaclara.core.constants as con
import aguaclara.core.physchem as pc
import aguaclara.core.pipes as pipe
import aguaclara.core.utility as ut
import aguaclara.core.drills as drills
from aguaclara.core.units import unit_registry as u

# from onshapepy.part import Part

import numpy as np
import math


class LFOM:

    safety_factor = 1.5
    sdr = 26
    drill_bits = drills.DRILL_BITS_D_IMPERIAL
    s_orfice = 1*u.cm
    # cad = Part(
    #     "https://cad.onshape.com/documents/e1798ab5f546e1414e86992d/w/104d463fef6c6a71c703abe6/e/890edb42c7884277d8d8711d"
    # )

    def __init__(self, q=20*u.L/u.s, hl=20*u.cm):
        self.q = q
        self.hl = hl

    def width_stout(self, z):
        """Return the width of a Stout weir at elevation z. More info
        here. <https://confluence.cornell.edu/display/AGUACLARA/LFOM+sutro+weir+research>
        """
        w_per_flow = 2 / ((2 * pc.gravity * z) ** (1 / 2) * con.VC_ORIFICE_RATIO * np.pi * self.hl)
        return w_per_flow*self.q

    @property
    def n_rows(self):
        """This equation states that the open area corresponding to one row
        can be set equal to two orifices of diameter=row height. If there
        are more than two orifices per row at the top of the LFOM then there
        are more orifices than are convenient to drill and more than
        necessary for good accuracy. Thus this relationship can be used to
        increase the spacing between the rows and thus increase the diameter
        of the orifices. This spacing function also sets the lower depth on
        the high flow rate LFOM with no accurate flows below a depth equal
        to the first row height.

        But it might be better to always set then number of rows to 10.
        The challenge is to figure out a reasonable system of constraints that
        reliably returns a valid solution.
        """
        N_estimated = (self.hl * np.pi / (2 * self.width_stout(self.hl) * self.q))
        variablerow = min(10, max(4, math.trunc(N_estimated.magnitude)))
        return variablerow

    @property
    def b_rows(self):
        """The distance center to center between each row of orifices.
        Message how long it took to load everything (minus packages)"""
        return self.hl / self.n_rows


    @property
    def vel_critical(self):
        """The average vertical velocity of the water inside the LFOM pipe
        at the very bottom of the bottom row of orifices The speed of
        falling water is 0.841 m/s for all linear flow orifice meters of
        height 20 cm, independent of total plant flow rate. """
        return 4 / (3 * math.pi) * (2 * pc.gravity * self.hl) ** (1 / 2)

    @property
    def area_pipe_min(self):
        """The minimum cross-sectional area of the LFOM pipe that assures a safety factor."""
        return self.safety_factor * self.q / self.vel_critical

    @property
    def nom_diam_pipe(self):
        """The nominal diameter of the LFOM pipe"""
        ID = pc.diam_circle(self.area_pipe_min)
        return pipe.ND_SDR_available(ID, self.sdr)

    @property
    def area_top_orifice(self):
        """Estimate the orifice area corresponding to the top row of orifices.
        Another solution method is to use integration to solve this problem.
        Here we use the width of the stout weir in the center of the top row
        to estimate the area of the top orifice
        """
        # Calculate the center of the top row:
        z = self.hl - 0.5 * self.b_rows
        # Multiply the stout weir width by the height of one row.
        return self.width_stout(z) * self.b_rows

    @property
    def d_orifice_max(self):
        """Determine the maximum orifice diameter."""
        return pc.diam_circle(self.area_top_orifice)

    @property
    def orifice_diameter(self):
        """The actual orifice diameter. We don't let the diameter extend beyond its row space. """
        maxdrill = min(self.b_rows, self.d_orifice_max)
        return ut.floor_nearest(maxdrill, self.drill_bits)

    @property
    def drillbit_area(self):
        """The area of the actual drill bit."""
        return pc.area_circle(self.orifice_diameter)

    @property
    def n_orifices_per_row_max(self):
        """A bound on the number of orifices allowed in each row.
        The distance between consecutive orifices must be enough to retain
        structural integrity of the pipe.
        """
        c = math.pi * pipe.ID_SDR(self.nom_diam_pipe, self.sdr)
        b = self.orifice_diameter + self.s_orfice

        return math.floor(c/b)

    @property
    def flow_ramp(self):
        """An equally spaced array representing flow at each row."""
        return np.linspace(self.q / self.n_rows, self.q, self.n_rows)*self.q.units

    @property
    def height_orifices(self):
        """Calculates the height of the center of each row of orifices.
        The bottom of the bottom row orifices is at the zero elevation
        point of the LFOM so that the flow goes to zero when the water height
        is at zero.
        """
        return np.arange((self.orifice_diameter * 0.5), self.hl, self.q)

    def flow_actual(self, Row_Index_Submerged, N_LFOM_Orifices):
        """Calculates the flow for a given number of submerged rows of orifices
        harray is the distance from the water level to the center of the orifices
        when the water is at the max level.

        Parameters
        ----------
        Row_Index_Submerged: int
            The index of the submerged row. All rows below and including this index are submerged.
        N_LFOM_Orifices: [int]
            The number of orifices at each row.

        Returns
        --------
        The flow through all of the orifices that are submerged.

        """
        harray = (np.linspace(self.b_rows, self.hl, self.n_rows))*self.hl.units - 0.5 * self.orifice_diameter
        FLOW_new = 0
        for i in range(Row_Index_Submerged + 1):
            FLOW_new = FLOW_new + (N_LFOM_Orifices[i] * (
                pc.flow_orifice_vert(self.orifice_diameter, harray[Row_Index_Submerged - i],
                                     con.VC_ORIFICE_RATIO)))
        return FLOW_new

    @property
    def n_orifices_per_row(self):
        """Calculate number of orifices at each level given a diameter.
        """
        # H is distance from the elevation between two rows of orifices down to the bottom of the orifices
        H = self.b_rows/2 + 0.5*self.orifice_diameter
        flow_per_orifice = pc.flow_orifice_vert(self.orifice_diameter, H, con.VC_ORIFICE_RATIO)
        n = np.zeros(self.n_rows)
        for i in range(self.n_rows):
            # calculate the ideal number of orifices at the current row without constraining to an integer
            flow_needed = self.flow_ramp[i] - self.flow_actual(i, n)
            n_orifices_real = (flow_needed / flow_per_orifice).to(u.dimensionless)
            # constrain number of orifices to be less than the max  per row and greater or equal to 0
            n[i] = min((max(0, round(n_orifices_real))), self.n_orifices_per_row_max)
        return n

    @property
    def error_per_row(self):
        """This function calculates the error of the design based on the differences between the predicted flow rate
        and the actual flow rate through the LFOM."""
        FLOW_lfom_error = []
        for i in range(self.n_rows-1):
            actual_flow = self.flow_actual(i, self.n_orifices_per_row)
            row_error = (actual_flow - self.flow_ramp[i]) / self.q
            FLOW_lfom_error.append(row_error.to(u.dimensionless))
        return FLOW_lfom_error

    # def draw(self):
    #     """Draw the LFOM in CAD."""
    #     self.cad.params = {"dHoles": self.orifice_diameter, "nHolesPerRow": str(self.n_orifices_per_row),
    #                        "OD": self.nom_diam_pipe, "bRows": self.b_rows}



import unittest
from aguaclara.core.units import unit_registry as u
import numpy as np
import aguaclara.core.pipeline as pipeline



class PipelineUtilityTest(unittest.TestCase):
    def test_pipeline_flow(self):
        flow = pipeline.flow_pipeline(np.array([3,4,5])*u.inch, np.array([3,4,5])*u.m, np.array([3,4,5]), 5 *u.m)
        self.assertAlmostEqual(flow.magnitude, 0.018149097841279497 * u.m**3 / u.s)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

setup(name='aguaclara',
      version='0.0.17',
      description='Open source functions for AguaClara water treatment research and plant design.',
      url='https://github.com/AguaClara/aguaclara',
      author='AguaClara at Cornell',
      author_email='aguaclara@cornell.edu',
      license='MIT',
      packages=find_packages(),
      install_requires=['pint','numpy','pandas','matplotlib','scipy'],
      setup_requires=["pytest-runner"],
      tests_require=["pytest"],
      include_package_data=True,
      test_suite="tests",
      zip_safe=False)
